import re

from django.http import JsonResponse
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.generics import get_object_or_404
from rest_framework.viewsets import ModelViewSet


class EditorModelMixin(object):
    pass


class DatatablesEditorModelViewSet(EditorModelMixin, ModelViewSet):

    @staticmethod
    def get_post_date(post):
        def read_date(data_in, data_out, rest_of_line):
            field_name = data_in[0]
            if not isinstance(data_out.get(field_name), dict):
                new_data_point = {}
                data_out[field_name] = new_data_point
            else:
                new_data_point = data_out[field_name]
            if len(data_in) == 2:
                new_data_point[data_in[1]] = rest_of_line
            else:
                read_date(data_in[1:], new_data_point, rest_of_line)

        data = {}
        for (line, value) in post.items():
            if line.startswith('data'):
                line_data = re.findall(r"\[([^\[\]]*)\]", line)
                read_date(line_data, data, value)
        return data

    @action(detail=False, url_name='editor', methods=['post'])
    def editor(self, request):
        post = request.POST
        act = post['action']
        data = self.get_post_date(post)

        return_data = []
        if act == 'edit' or act == 'remove' or act == 'create':
            for elem_id, changes in data.items():
                if act == 'create':
                    serializer = self.serializer_class(
                        data=changes,
                        context={'request': request}
                    )
                    if not serializer.is_valid():  # pragma: no cover
                        raise ValidationError(serializer.errors)
                    serializer.save()
                    return_data.append(serializer.data)
                    continue

                elem = get_object_or_404(self.queryset, pk=elem_id)
                if act == 'edit':
                    serializer = self.serializer_class(
                        instance=elem, data=changes,
                        partial=True, context={'request': request}
                    )
                    if not serializer.is_valid():  # pragma: no cover
                        raise ValidationError(serializer.errors)
                    serializer.save()
                    return_data.append(serializer.data)
                elif act == 'remove':
                    elem.delete()

        return JsonResponse({'data': return_data})



from collections import OrderedDict

from django.core.paginator import InvalidPage
from django.utils import six
from rest_framework.exceptions import NotFound
from rest_framework.pagination import (
    PageNumberPagination, LimitOffsetPagination
)
from rest_framework.response import Response


class DatatablesMixin(object):
    def get_paginated_response(self, data):
        if not self.is_datatable_request:
            return super(DatatablesMixin, self).get_paginated_response(data)

        return Response(OrderedDict([
            ('recordsTotal', self.total_count),
            ('recordsFiltered', self.count),
            ('data', data)
        ]))

    def get_count_and_total_count(self, queryset, view):
        if hasattr(view, '_datatables_filtered_count'):
            count = view._datatables_filtered_count
            del view._datatables_filtered_count
        else:  # pragma: no cover
            count = queryset.count()
        if hasattr(view, '_datatables_total_count'):
            total_count = view._datatables_total_count
            del view._datatables_total_count
        else:  # pragma: no cover
            total_count = count
        return count, total_count


class DatatablesPageNumberPagination(DatatablesMixin, PageNumberPagination):
    def paginate_queryset(self, queryset, request, view=None):
        if request.accepted_renderer.format != 'datatables':
            self.is_datatable_request = False
            return super(
                DatatablesPageNumberPagination, self
            ).paginate_queryset(queryset, request, view)
        if request.query_params.get('length') is None:
            return None
        self.count, self.total_count = self.get_count_and_total_count(
            queryset, view
        )
        self.is_datatable_request = True
        self.page_size_query_param = 'length'
        page_size = self.get_page_size(request)
        if not page_size:  # pragma: no cover
            return None

        paginator = self.django_paginator_class(queryset, page_size)
        start = int(request.query_params.get('start', 0))
        page_number = int(start / page_size) + 1

        try:
            self.page = paginator.page(page_number)
        except InvalidPage as exc:
            msg = self.invalid_page_message.format(
                page_number=page_number, message=six.text_type(exc)
            )
            raise NotFound(msg)
        self.request = request
        return list(self.page)


class DatatablesLimitOffsetPagination(DatatablesMixin, LimitOffsetPagination):
    def paginate_queryset(self, queryset, request, view=None):
        if request.accepted_renderer.format == 'datatables':
            self.is_datatable_request = True
            if request.query_params.get('length') is None:
                return None
            self.limit_query_param = 'length'
            self.offset_query_param = 'start'
            self.count, self.total_count = self.get_count_and_total_count(
                queryset, view
            )
        else:
            self.is_datatable_request = False
        return super(
            DatatablesLimitOffsetPagination, self
        ).paginate_queryset(queryset, request, view)



import re
from copy import deepcopy

from django.db.models import Q

from rest_framework.filters import BaseFilterBackend


class DatatablesFilterBackend(BaseFilterBackend):
    """
    Filter that works with datatables params.
    """
    def filter_queryset(self, request, queryset, view):
        if request.accepted_renderer.format != 'datatables':
            return queryset

        filtered_count_before = queryset.count()
        total_count = view.get_queryset().count()
        # set the queryset count as an attribute of the view for later
        # TODO: find a better way than this hack
        setattr(view, '_datatables_total_count', total_count)

        # parse query params
        getter = request.query_params.get
        fields = self.get_fields(getter)
        ordering = self.get_ordering(getter, fields)
        search_value = getter('search[value]')
        search_regex = getter('search[regex]') == 'true'

        # filter queryset
        q = Q()
        for f in fields:
            if not f['searchable']:
                continue
            if search_value and search_value != 'false':
                if search_regex:
                    if self.is_valid_regex(search_value):
                        # iterate through the list created from the 'name'
                        # param and create a string of 'ior' Q() objects.
                        for x in f['name']:
                            q |= Q(**{'%s__iregex' % x: search_value})
                else:
                    # same as above.
                    for x in f['name']:
                        q |= Q(**{'%s__icontains' % x: search_value})
            f_search_value = f.get('search_value')
            f_search_regex = f.get('search_regex') == 'true'
            if f_search_value:
                if f_search_regex:
                    if self.is_valid_regex(f_search_value):
                        # create a temporary q variable to hold the Q()
                        # objects adhering to the field's name criteria.
                        temp_q = Q()
                        for x in f['name']:
                            temp_q |= Q(**{'%s__iregex' % x: f_search_value})
                        # Use deepcopy() to transfer them to the global Q()
                        # object. Deepcopy() necessary, since the var will be
                        # reinstantiated next iteration.
                        q = q & deepcopy(temp_q)
                else:
                    temp_q = Q()
                    for x in f['name']:
                        temp_q |= Q(**{'%s__icontains' % x: f_search_value})
                    q = q & deepcopy(temp_q)

        if q:
            queryset = queryset.filter(q).distinct()
            filtered_count = queryset.count()
        else:
            filtered_count = filtered_count_before
        # set the queryset count as an attribute of the view for later
        # TODO: maybe find a better way than this hack ?
        setattr(view, '_datatables_filtered_count', filtered_count)

        # order queryset
        if len(ordering):
            queryset = queryset.order_by(*ordering)
        return queryset

    def get_fields(self, getter):
        fields = []
        i = 0
        while True:
            col = 'columns[%d][%s]'
            data = getter(col % (i, 'data'))
            if data is None or not data:
                break
            name = getter(col % (i, 'name'))
            if not name:
                name = data
            search_col = col % (i, 'search')
            # to be able to search across multiple fields (e.g. to search
            # through concatenated names), we create a list of the name field,
            # replacing dot notation with double-underscores and splitting
            # along the commas.
            field = {
                'name': [
                    n.lstrip() for n in name.replace('.', '__').split(',')
                ],
                'data': data,
                'searchable': getter(col % (i, 'searchable')) == 'true',
                'orderable': getter(col % (i, 'orderable')) == 'true',
                'search_value': getter('%s[%s]' % (search_col, 'value')),
                'search_regex': getter('%s[%s]' % (search_col, 'regex')),
            }
            fields.append(field)
            i += 1
        return fields

    def get_ordering(self, getter, fields):
        ordering = []
        i = 0
        while True:
            col = 'order[%d][%s]'
            idx = getter(col % (i, 'column'))
            if idx is None:
                break
            try:
                field = fields[int(idx)]
            except IndexError:
                i += 1
                continue
            if not field['orderable']:
                i += 1
                continue
            dir_ = getter(col % (i, 'dir'), 'asc')
            ordering.append('%s%s' % (
                '-' if dir_ == 'desc' else '',
                field['name'][0]
            ))
            i += 1
        return ordering

    def is_valid_regex(cls, regex):
        try:
            re.compile(regex)
            return True
        except re.error:
            return False
